from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.non_deterministic_finite_automaton import NondeterministicFiniteAutomaton as NFA
from pythautomata.model_comparators.hopcroft_karp_comparison_strategy import HopcroftKarpComparisonStrategy as HopcroftKarpComparison
from pythautomata.utilities.automata_converter import AutomataConverter

abAlphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'))))
abcAlphabet = Alphabet(
    frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr('c'))))

# TODO: DOCUMENT AND TEST


class SampleNFAs:
    """
    Class containing sample NFAs

    Methods
    -------

    """
    @staticmethod
    def get_all_automata():
        return [
            SampleNFAs.get_more_membership_queries_automaton(),
            SampleNFAs.get_more_equivalence_queries_automaton(),
            SampleNFAs.get_state_count_does_not_increase_automaton(),
            SampleNFAs.get_intermediate_hypothesis_automaton(),
            SampleNFAs.get_algorithm_would_not_terminate_automaton(),
            SampleNFAs.get_evolution_of_the_measure_automaton()
        ]

    # Σ*aΣ
    @staticmethod
    def get_more_membership_queries_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4")

        state0.add_transition(a, state1)
        state0.add_transition(b, state1)
        state1.add_transition(a, state2)
        state1.add_transition(b, state2)
        state2.add_transition(a, state3)
        state2.add_transition(b, state3)
        state3.add_transition(a, state4)
        state3.add_transition(b, state0)
        state4.add_transition(a, state4)
        state4.add_transition(b, state4)
        comparator = HopcroftKarpComparison()
        return NFA(abAlphabet, frozenset({state0}),
                   set([state0, state1,
                        state2, state3, state4]),
                   comparator,
                   "Paper - More membership queries automaton")

    @staticmethod
    def get_more_equivalence_queries_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4", True)
        state5 = State("State 5")

        state0.add_transition(a, state2)
        state0.add_transition(b, state1)
        state1.add_transition(a, state3)
        state1.add_transition(b, state1)
        state2.add_transition(a, state2)
        state2.add_transition(b, state4)
        state3.add_transition(a, state5)
        state3.add_transition(b, state5)
        state4.add_transition(a, state5)
        state4.add_transition(b, state4)
        state5.add_transition(a, state5)
        state5.add_transition(b, state5)
        comparator = HopcroftKarpComparison()
        return NFA(abAlphabet, frozenset({state0}),
                   set([state0, state1, state2,
                        state3, state4, state5]),
                   comparator,
                   "Paper - More equivalence queries automaton")

    @staticmethod
    def get_state_count_does_not_increase_automaton():
        a = abcAlphabet['a']
        b = abcAlphabet['b']
        c = abcAlphabet['c']

        state0 = State("State 0", True)
        state1 = State("State 1")
        state2 = State("State 2", True)
        state3 = State("State 3")

        state0.add_multiple_transitions(a, [state0, state1, state2])
        state0.add_multiple_transitions(b, [state1, state3])
        state0.add_transition(c, state3)

        state1.add_transition(a, state2)
        state1.add_multiple_transitions(b, [state0, state1, state2])
        state1.add_multiple_transitions(c, [state1, state3])

        state2.add_transition(a, state1)
        state2.add_transition(b, state3)
        state2.add_transition(c, state3)

        state3.add_transition(a, state2)
        state3.add_transition(b, state3)
        state3.add_transition(c, state2)

        comparator = HopcroftKarpComparison()
        result = NFA(abcAlphabet, frozenset({state0, state2}),
                     set([state0, state1, state2, state3]),
                     comparator,
                     "Paper - State count does not increase automaton")
        result._queryable_self = AutomataConverter.convert_nfa_to_dfa(result)
        return result

    @staticmethod
    def get_intermediate_hypothesis_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("State 0")
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3")
        state4 = State("State 4")

        state0.add_transition(a, state4)
        state0.add_transition(b, state1)
        state1.add_transition(a, state2)
        state1.add_transition(b, state1)
        state2.add_transition(a, state4)
        state2.add_transition(b, state3)
        state3.add_transition(a, state2)
        state3.add_transition(b, state2)
        state4.add_transition(a, state4)
        state4.add_transition(b, state4)
        comparator = HopcroftKarpComparison()
        return NFA(abAlphabet, frozenset({state0}),
                   set([state0, state1, state2, state3, state4]),
                   comparator,
                   "Paper - Intermediate hypothesis automaton")

    @staticmethod
    def get_algorithm_would_not_terminate_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2")
        state3 = State("State 3")
        state4 = State("State 4")
        state5 = State("State 5")
        state6 = State("State 6", True)
        state7 = State("State 7")
        state8 = State("State 8", True)
        state9 = State("State 9", True)

        state0.add_transition(a, state1)
        state0.add_transition(b, state9)
        state1.add_transition(a, state2)
        state1.add_transition(b, state7)
        state2.add_transition(a, state3)
        state3.add_transition(a, state2)
        state3.add_transition(b, state4)
        state4.add_transition(b, state5)
        state5.add_transition(b, state6)
        state7.add_transition(b, state8)
        state8.add_transition(b, state6)
        state9.add_transition(b, state9)
        comparator = HopcroftKarpComparison()
        return NFA(abAlphabet, frozenset({state0}),
                   set([state0, state1, state2, state3, state4,
                        state5, state6, state7, state8, state9]),
                   comparator,
                   "Paper - Algorithm would not terminate automaton")

    @staticmethod
    def get_evolution_of_the_measure_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3", True)
        state4 = State("State 4")
        state5 = State("State 5", True)
        state6 = State("State 6")

        state0.add_transition(b, state1)
        state1.add_transition(a, state2)
        state2.add_transition(a, state3)
        state3.add_transition(a, state5)
        state3.add_transition(b, state4)
        state4.add_transition(a, state5)
        comparator = HopcroftKarpComparison()
        return NFA(abAlphabet, frozenset({state0}),
                   set([state0, state1, state2,
                        state3, state4, state5, state6]),
                   comparator,
                   "Paper - Evolution of the measure automaton")



import uuid
import numpy as np

from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from itertools import chain
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from decimal import Decimal

from pythautomata.abstract.pdfa_model_exporting_strategy import PDFAModelExportingStrategy
from pythautomata.model_exporters.wfa_image_exporter import WFAImageExporter
from pythautomata.base_types.symbol import Symbol
from pythautomata.abstract.finite_automaton import FiniteAutomataComparator, FiniteAutomaton
from typing import Any

epsilon = Sequence()


class WeightedAutomaton(FiniteAutomaton):

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def terminal_symbol(self):
        return self._terminal_symbol

    @terminal_symbol.setter
    def terminal_symbol(self, value):
        self._terminal_symbol = value

    @property
    def alphabet(self):
        return self._alphabet

    @alphabet.setter
    def alphabet(self, value):
        self._alphabet = value

    def __init__(self, alphabet: Alphabet, weighted_states: set, terminal_symbol: Symbol,
                 comparator: FiniteAutomataComparator, name=None,
                 export_strategies: list[PDFAModelExportingStrategy] = None):
        super().__init__(comparator)
        if export_strategies is None:
            export_strategies = [WFAImageExporter()]
        if name is None:
            self.name = 'WFA - ' + str(uuid.uuid4().hex)
        else:
            self._name = name
        self._terminal_symbol = terminal_symbol
        self._alphabet = alphabet
        self.weighted_states = weighted_states
        self._exporting_strategies = export_strategies

    @property
    def hole(self):
        return None

    @property
    def initial_states(self) -> frozenset:
        initial_states = list(
            filter(lambda state: state.initial_weight != 0, self.weighted_states))
        return frozenset(initial_states)

    def sequence_weight(self, sequence: Sequence):
        return float(sum(map(lambda x: self._sequence_weight_from(sequence.value, x, x.initial_weight),
                             self.weighted_states)))

    def _sequence_weight_from(self, sequence_value, weighted_state: WeightedState, weight):
        if weight == 0:
            return Decimal(0)
        else:
            if not sequence_value:
                return Decimal(weight)
            else:
                if sequence_value[0] == self.terminal_symbol:
                    return Decimal(weight * weighted_state.final_weight)
                else:
                    transitions = weighted_state.transitions_list_for(
                        sequence_value[0])
                    transitions_unzipped = list(zip(*transitions))
                    next_states = transitions_unzipped[0]
                    weights = transitions_unzipped[1]
                    return sum(map(lambda x, y: Decimal(weight) * self._sequence_weight_from(sequence_value[1:], x, y),
                                   next_states, weights))

    def log_sequence_weight(self, sequence: Sequence):
        return float(sum(filter(lambda x: x != float('inf'),
                                map(lambda x: self._log_sequence_weight_from(sequence.value, x, x.initial_weight),
                                    self.weighted_states))))

    def _log_sequence_weight_from(self, sequence_value, weighted_state: WeightedState, weight):
        if weight == 0:
            return float('inf')
        else:
            if not sequence_value:
                return np.log(weight)
            else:
                if sequence_value[0] == self.terminal_symbol:
                    return np.log(weight) + np.log(weighted_state.final_weight)
                else:
                    transitions = weighted_state.transitions_list_for(
                        sequence_value[0])
                    transitions_unzipped = list(zip(*transitions))
                    next_states = transitions_unzipped[0]
                    weights = transitions_unzipped[1]
                    return sum(map(lambda x, y: np.log(weight) +
                                   self._log_sequence_weight_from(
                                       sequence_value[1:], x, y),
                                   next_states, weights))

    def last_token_weight(self, sequence: Sequence):
        return list(chain.from_iterable([self._last_token_weight_from(sequence.value, y, y.initial_weight)
                                         for y in self.weighted_states if y.initial_weight > 0]))

    def _last_token_weight_from(self, sequence_value, state: WeightedState, weight):
        # if weight == 0:
        #    return [0]
        if not sequence_value:
            return [weight]
        else:
            if sequence_value[0] == self.terminal_symbol:
                return [state.final_weight]
            else:
                transitions = state.transitions_list_for(sequence_value[0])
                transitions_unzipped = list(zip(*transitions))
                next_states = transitions_unzipped[0]
                weights = transitions_unzipped[1]
                return chain.from_iterable(map(lambda x, y: self._last_token_weight_from(sequence_value[1:], x, y),
                                               next_states, weights))

    def get_last_token_weights(self, sequence, required_suffixes):
        weights = list()
        for suffix in required_suffixes:
            new_seq = sequence + suffix
            weight = self.last_token_weight(new_seq)
            if len(weight) > 0:
                weights.append(weight[0])
            else:
                weights.append(0)
        return weights

    def get_last_token_weights_batch(self, sequences, required_suffixes):
        result = []
        for sequence in sequences:
            result.append(self.get_last_token_weights(
                sequence, required_suffixes))
        return result

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, WeightedAutomaton) and self._comparator.are_equivalent(self, other)




from typing import Optional

from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.state import State


class MooreMachineComparisonStrategy():

    def are_equivalent(self, automaton1, automaton2) -> bool:
        return self.get_counterexample_between(automaton1, automaton2) is None

    def get_counterexample_between(self, mma1: MooreMachineAutomaton, mma2: MooreMachineAutomaton) -> Optional[Sequence]:
        if mma1._input_alphabet != mma2._input_alphabet:
            raise ValueError("Input alphabets are not equivalent.")
        if mma1._output_alphabet != mma2._output_alphabet:
            raise ValueError("Output alphabets are not equivalent.")

        # initialPairs is an arbitrary pair of states
        initialPair: tuple[State, State] = (
            mma1.initial_state, mma2.initial_state)
        pairsToVisit = [initialPair]
        sequenceForPairs = {initialPair: Sequence()}
        visitedPairs: set[tuple[State, State]] = set()

        while pairsToVisit:
            pair = pairsToVisit[0]
            if pair[0].value != pair[1].value:
                return sequenceForPairs[pair]
            for symbol in mma1._input_alphabet.symbols:
                self._process_equivalence_iteration_with(symbol, pairsToVisit,
                                                         visitedPairs, sequenceForPairs)
            pairsToVisit.remove(pair)
            visitedPairs.add(pair)
        return None

    def _process_equivalence_iteration_with(self, symbol, pairs_to_visit, visited_pairs, sequence_for_pairs):
        pair = pairs_to_visit[0]
        selfNextState = min(pair[0].next_states_for(symbol))
        otherNextState = min(pair[1].next_states_for(symbol))
        nextPair = (selfNextState, otherNextState)
        if nextPair not in pairs_to_visit and nextPair not in visited_pairs:
            sequence_for_pairs[nextPair] = sequence_for_pairs[pair] + symbol
            pairs_to_visit.append(nextPair)

    def equivalent_output(self, observation1, observation2) -> bool:
        return observation1 == observation2



from re import findall, match
from typing import Any

from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.exceptions.model_importing_error import ModelImportingError


class EncodedStringImporter:

    @staticmethod
    def import_automata_attributes(code):
        fileLines = str.split(code, '\n')
        modelName = fileLines[0]
        alphabet = EncodedStringImporter.__get_alphabet_from_data(
            fileLines[1:4])
        del fileLines[:4]
        statesByName, initialStates, currentIndex = EncodedStringImporter.__get_states_from_data(
            fileLines)
        del fileLines[:currentIndex + 3]
        EncodedStringImporter.__get_transitions_from_data(
            fileLines, alphabet, statesByName)

        return alphabet, initialStates, set(statesByName.values()), modelName

    @staticmethod
    def __get_alphabet_from_data(fileLines: list) -> Alphabet:
        if not match(r"^(?:Alphabet: )\[(?:'([^']+)')(?:, (?:'([^']+)')+)*\]$", fileLines[0]):
            raise ModelImportingError()
        return Alphabet(set(SymbolStr(symbol)
                            for symbol in findall(r"(?:'([^']+)')", fileLines[0])))

    @staticmethod
    def __get_states_from_data(fileLines: list) -> tuple[dict[Any, State], frozenset[State], int]:
        statesByName = {}
        initialStates = set()
        currentIndex = 0

        while not fileLines[currentIndex].isspace() and not fileLines[currentIndex] == '':
            currentLine = fileLines[currentIndex]
            stateData = match(
                r"^\t(?:\"([^\"]+)\")( \(Initial\))?( \(Final\))?$", currentLine)

            if not stateData:
                raise ModelImportingError()

            name = stateData.group(1)
            isInitial = stateData.group(2) is not None
            isFinal = stateData.group(3) is not None

            state = State(name, isFinal)
            statesByName[name] = state
            if isInitial:
                initialStates.add(state)
            currentIndex += 1
        return statesByName, frozenset(initialStates), currentIndex

    @staticmethod
    def __get_transitions_from_data(fileLines: list, alphabet: Alphabet, statesByName: dict):
        currentIndex = 0
        while True:
            originName = match(
                r"^(?:\"([^\"]+)\"):$", fileLines[currentIndex])
            origin = statesByName[originName.group(1)]
            currentIndex += 1

            while True:
                if currentIndex >= len(fileLines):
                    return

                currentLine = fileLines[currentIndex]
                transitionData = match(
                    r"^\t(?:'([^']+)')(?: -> )\[(?:\"([^\"]+)\")(?:, (?:\"([^\"]+)\"))*\]$", currentLine)
                if not transitionData:
                    break
                symbol = alphabet[transitionData.group(1)]
                destinations = set(statesByName[name]
                                   for name in findall(r"(?:\"([^\"]+)\")", currentLine))
                origin.transitions[symbol] = destinations
                currentIndex += 1



from .python_automata_exception import PythonAutomataException

class UnknownSymbolsException(PythonAutomataException):
    def __init__(self):
        self.message = "Some symbols do not belong to the given alphabet"


import uuid
from pythautomata.abstract.boolean_model import BooleanModel

from pythautomata.base_types.state import State
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import Symbol
from pythautomata.abstract.finite_automaton import FiniteAutomaton, FiniteAutomataComparator
from pythautomata.exceptions.unknown_symbols_exception import UnknownSymbolsException
from pythautomata.exceptions.non_deterministic_states_exception import NonDeterministicStatesException
from pythautomata.model_exporters.encoded_file_exporting_strategy import EncodedFileExportingStrategy
from pythautomata.model_exporters.encoded_string_exporting_strategy import EncodedStringExportingStrategy
from pythautomata.utilities.encoded_string_importer import EncodedStringImporter
from pythautomata.utilities.encoded_file_importer import EncodedFileImporter


class DeterministicFiniteAutomaton(FiniteAutomaton, BooleanModel):
    """
    Implementation of DFA.

    Attributes
    ----------
    states: set[State]
        Set containing the DFA's states
    initial_state: State
        Initial state of the DFA. Also included in "states"
    """

    def __init__(self, alphabet: Alphabet, initial_state: State, states: set[State],
                 comparator: FiniteAutomataComparator, name: str = None,
                 exportingStrategies: list = [EncodedFileExportingStrategy()], hole: State = State("Hole")):
        self.states = states
        for state in self.states:
            self._verify_state(state, alphabet)
            state.add_hole_transition(hole)

        self._name = 'DFA - ' + str(uuid.uuid4().hex) if name is None else name
        self._alphabet = alphabet
        self.initial_state = initial_state
        self._set_hole(hole)
        self._exporting_strategies = exportingStrategies
        self._actual_state = self.initial_state
        super(DeterministicFiniteAutomaton, self).__init__(comparator)

    def accepts(self, sequence: Sequence) -> bool:
        actual_state = self.initial_state
        for symbol in sequence.value:
            actual_state = actual_state.next_state_for(symbol)
        return actual_state.is_final

    @property
    def initial_states(self) -> frozenset:
        return frozenset([self.initial_state])

    @property
    def hole(self):
        return self._hole

    def step(self, symbol):
        self._actual_state = self._actual_state.next_state_for(symbol)
        return self._actual_state.is_final

    def reset(self):
        self._actual_state = self.initial_state

    def clone(self) -> 'DeterministicFiniteAutomaton':
        return DeterministicFiniteAutomaton(self.alphabet,
                                            self.initial_state,
                                            self.states,
                                            self.comparator,
                                            self.name)

    def _set_hole(self, hole: State) -> None:
        self._hole = hole
        # hole's hole state is itself
        self._hole.add_hole_transition(self.hole)

    def _verify_state(self, state: State, alphabet: Alphabet) -> None:
        self._verify_transition_symbols_in_alphabet(
            state.transitions, alphabet)
        self._verify_state_is_deterministic(state)

    def _verify_transition_symbols_in_alphabet(self, transitions: dict[Symbol, set['State']], alphabet: Alphabet) -> \
            None:
        if not all(symbol in alphabet for symbol in transitions):
            raise UnknownSymbolsException()

    def _verify_state_is_deterministic(self, state: State) -> None:
        if not state.is_deterministic:
            raise NonDeterministicStatesException()

    def __getstate__(self):
        # dump a string
        lines = EncodedStringExportingStrategy().export(self)
        return lines

    def __setstate__(self, state):
        attrs = EncodedStringImporter().import_automata_attributes(str(state))
        alphabet, initialStates, states, modelName = attrs
        self.name = modelName
        self._alphabet = alphabet
        self.initial_state = list(initialStates)[0]
        self.states = states
        hole = State("Hole")
        self._set_hole(hole)
        for state in self.states:
            assert all(
                symbol in self.alphabet for symbol in state.transitions.keys())
            state.add_hole_transition(self.hole)
        self._queryable_self = self



from unittest import TestCase

from pythautomata.base_types.sequence import Sequence
from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.utilities.automata_converter import AutomataConverter
from pythautomata.model_comparators.moore_machine_comparison_strategy import MooreMachineComparisonStrategy
from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars as Tomitas

class TestDFAToMooreMachine(TestCase):
    def test_dfa_to_moore_machine_tomitas_automaton_1(self):
        moore_machine = SampleMooreMachines.get_tomitas_automaton_1()
        dfa = Tomitas.get_automaton_1()
        converted_dfa = AutomataConverter.convert_dfa_to_moore_machine(dfa)

        return self.assertEqual(moore_machine, converted_dfa)
    
    def test_dfa_to_moore_machine_tomitas_automaton_2(self):
        moore_machine = SampleMooreMachines.get_tomitas_automaton_2()
        dfa = Tomitas.get_automaton_2()
        converted_dfa = AutomataConverter.convert_dfa_to_moore_machine(dfa)

        return self.assertEqual(moore_machine, converted_dfa)





from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton as DFA
from pythautomata.automata.non_deterministic_finite_automaton import NondeterministicFiniteAutomaton as NFA
from pythautomata.model_comparators.hopcroft_karp_comparison_strategy import HopcroftKarpComparisonStrategy as HopcroftKarpComparison
from pythautomata.model_comparators.dfa_comparison_strategy import DFAComparisonStrategy as DFAComparator

binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
zero = binaryAlphabet['0']
one = binaryAlphabet['1']

abAlphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'))))
abcAlphabet = Alphabet(
    frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr('c'))))
a = abcAlphabet['a']
b = abcAlphabet['b']
c = abcAlphabet['c']

abcdAlphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'),
                                   SymbolStr('c'), SymbolStr('d'))))

abcdeAlphabet = Alphabet(
    frozenset((a, b, c, SymbolStr('d'), SymbolStr('e'))))
d = abcdeAlphabet['d']
e = abcdeAlphabet['e']

#TODO: DOCUMENT


class OtherAutomata:

    @staticmethod
    def get_all_automata():
        return [
            OtherAutomata.get_automaton_1(),
            OtherAutomata.get_automaton_1_minimized(),
            OtherAutomata.get_automaton_2(),
            OtherAutomata.get_automaton_2_minimized(),
            OtherAutomata.get_automaton_3(),
            OtherAutomata.get_automaton_3_minimized(),
            OtherAutomata.get_automaton_4(),
            OtherAutomata.get_automaton_4_minimized(),
            OtherAutomata.get_non_minimizable_automaton_1(),
            OtherAutomata.get_non_minimizable_automaton_2(),
            OtherAutomata.get_non_minimizable_automaton_3(),
            OtherAutomata.get_dfa_1(),
            OtherAutomata.get_nfa_1(),
            OtherAutomata.get_dfa_2(),
            OtherAutomata.get_nfa_2(),
            OtherAutomata.get_dfa_3(),
            OtherAutomata.get_nfa_3(),
            OtherAutomata.get_dfa_4(),
            OtherAutomata.get_nfa_4(),
            OtherAutomata.get_dfa_5(),
            OtherAutomata.get_nfa_5(),
            OtherAutomata.get_program_workflow_automaton(),
            OtherAutomata.get_ecommerce_automaton(),
            OtherAutomata.get_reduced_ecommerce_automaton(),
            OtherAutomata.get_dolzhenko_jonoska_automaton(),
            OtherAutomata.get_a_or_b_automaton(),
            OtherAutomata.get_ab_with_cs_automaton(),
            OtherAutomata.get_ab_automaton(),
            OtherAutomata.get_a_ending_automaton(),
            OtherAutomata.get_a_ending_with_cs_automaton(),
            OtherAutomata.get_zero_ending_automaton(),
            OtherAutomata.get_nonempty_zero_ending_automaton(),
            OtherAutomata.get_nonempty_one_ending_automaton(),
            OtherAutomata.get_simpler_ab_prefixed_automaton(),
            OtherAutomata.get_complex_ab_prefixed_automaton(),
            OtherAutomata.get_alternating_bit_protocol_automaton(),
            OtherAutomata.get_alternating_bit_protocol_z_automaton(),
            OtherAutomata.get_empty_automaton(),
            OtherAutomata.get_sigma_star_automaton(),
            OtherAutomata.get_uneven_number_of_as_automaton(),
            OtherAutomata.get_uneven_number_of_symbols_automaton(),
            OtherAutomata.get_uneven_number_of_as_and_symbols_automaton(),
            OtherAutomata.get_uneven_number_of_as_or_symbols_automaton()
        ]

    @staticmethod
    def get_automaton_1():
        state0 = State("State 0")
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3")
        state4 = State("State 4", True)
        state5 = State("State 5")

        state0.add_transition(zero, state3)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state2)
        state1.add_transition(one, state5)
        state2.add_transition(zero, state2)
        state2.add_transition(one, state5)
        state3.add_transition(zero, state0)
        state3.add_transition(one, state4)
        state4.add_transition(zero, state2)
        state4.add_transition(one, state5)
        state5.add_transition(zero, state5)
        state5.add_transition(one, state5)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state2, state3,
                        state4, state5]), comparator, "Automaton 1")

    @staticmethod
    def get_automaton_1_minimized():
        state0 = State("Part 0")
        state1 = State("Part 1", True)
        state2 = State("Part 2")
        state0.add_transition(zero, state0)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state1)
        state1.add_transition(one, state2)
        state2.add_transition(zero, state2)
        state2.add_transition(one, state2)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state2]), comparator, "Automaton 1 minimized")

    @staticmethod
    def get_automaton_2():
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state3 = State("State 3")
        state0.add_transition(zero, state3)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state2)
        state3.add_transition(zero, state2)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state2, state3]), comparator, "Automaton 2")

    @staticmethod
    def get_automaton_2_minimized():
        state0 = State("State 0")
        state1_and_state3 = State("State 1 and 3")
        state2 = State("State 2", True)
        state0.add_transition(zero, state1_and_state3)
        state0.add_transition(one, state1_and_state3)
        state1_and_state3.add_transition(zero, state2)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1_and_state3, state2]), comparator, "Automaton 2 minimized")

    @staticmethod
    def get_automaton_3():
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4")
        state5 = State("State 5")
        state6 = State("State 6")
        state7 = State("State 7")

        state1.add_transition(zero, state3)
        state1.add_transition(one, state2)
        state2.add_transition(zero, state3)
        state2.add_transition(one, state1)
        state3.add_transition(zero, state4)
        state3.add_transition(one, state2)
        state4.add_transition(one, state5)
        state5.add_transition(zero, state7)
        state5.add_transition(one, state6)
        state6.add_transition(one, state3)
        state7.add_transition(zero, state6)
        state7.add_transition(one, state6)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state1,
                   set([state1, state2, state3, state4, state5,
                        state6, state7]), comparator, "Automaton 3")

    @staticmethod
    def get_automaton_3_minimized():
        state1_and_state2 = State("State 1 and State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4")
        state5 = State("State 5")
        state6 = State("State 6")
        state7 = State("State 7")

        state1_and_state2.add_transition(zero, state3)
        state1_and_state2.add_transition(one, state1_and_state2)
        state3.add_transition(zero, state4)
        state3.add_transition(one, state1_and_state2)
        state4.add_transition(one, state5)
        state5.add_transition(zero, state7)
        state5.add_transition(one, state6)
        state6.add_transition(one, state3)
        state7.add_transition(zero, state6)
        state7.add_transition(one, state6)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state1_and_state2,
                   set([state1_and_state2, state3, state4, state5,
                        state6, state7]), comparator, "Automaton 3 minimized")

    @staticmethod
    def get_automaton_4():
        state_i = State("State i", True)
        state_d = State("State d")
        state_c = State("State c", True)
        state_a = State("State a")
        state_b = State("State b")
        state_g = State("State g")
        state_e = State("State e")
        state_h = State("State h")
        state_f = State("State f")

        state_i.add_transition(zero, state_i)
        state_i.add_transition(one, state_d)
        state_d.add_transition(zero, state_c)
        state_d.add_transition(one, state_e)
        state_c.add_transition(zero, state_a)
        state_c.add_transition(one, state_c)
        state_a.add_transition(zero, state_b)
        state_a.add_transition(one, state_f)
        state_b.add_transition(zero, state_g)
        state_b.add_transition(one, state_c)
        state_g.add_transition(zero, state_g)
        state_g.add_transition(one, state_e)
        state_e.add_transition(zero, state_h)
        state_e.add_transition(one, state_f)
        state_h.add_transition(zero, state_c)
        state_h.add_transition(one, state_g)
        state_f.add_transition(zero, state_c)
        state_f.add_transition(one, state_g)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state_i,
                   set([state_i, state_d, state_c, state_a, state_b,
                        state_g, state_e, state_h, state_f]), comparator,
                   "Automaton 4")

    @staticmethod
    def get_automaton_4_minimized():
        state_i = State("State i", True)
        state_d = State("State d")
        state_c = State("State c", True)
        state_a = State("State a")
        state_b = State("State b")
        state_g = State("State g")
        state_e = State("State e")
        state_h_and_f = State("State h and f")

        state_i.add_transition(zero, state_i)
        state_i.add_transition(one, state_d)
        state_d.add_transition(zero, state_c)
        state_d.add_transition(one, state_e)
        state_c.add_transition(zero, state_a)
        state_c.add_transition(one, state_c)
        state_a.add_transition(zero, state_b)
        state_a.add_transition(one, state_h_and_f)
        state_b.add_transition(zero, state_g)
        state_b.add_transition(one, state_c)
        state_g.add_transition(zero, state_g)
        state_g.add_transition(one, state_e)
        state_e.add_transition(zero, state_h_and_f)
        state_e.add_transition(one, state_h_and_f)
        state_h_and_f.add_transition(zero, state_c)
        state_h_and_f.add_transition(one, state_g)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state_i,
                   set([state_i, state_d, state_c, state_a, state_b, state_g,
                        state_e, state_h_and_f]), comparator, "Automaton 4 minimized")

    @staticmethod
    def get_non_minimizable_automaton_1():
        state0 = State("State 0", True)
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3")

        state0.add_transition(zero, state1)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state2)
        state1.add_transition(one, state0)
        state2.add_transition(zero, state1)
        state2.add_transition(one, state3)
        state3.add_transition(zero, state0)
        state3.add_transition(one, state1)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state2, state3]), comparator, "Non-minimizable automaton 1")

    @staticmethod
    def get_non_minimizable_automaton_2():
        state0 = State("State 0")
        state1 = State("State 1", True)
        state0.add_transition(zero, state1)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state0)
        state1.add_transition(one, state0)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1]), comparator, "Non-minimizable automaton 2")

    @staticmethod
    def get_non_minimizable_automaton_3():
        state0 = State("State 0")
        state1 = State("State 1", True)
        state2 = State("State 2")
        state0.add_transition(zero, state1)
        state0.add_transition(one, state2)
        state1.add_transition(zero, state1)
        state1.add_transition(one, state1)
        state2.add_transition(one, state0)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state2]), comparator, "Non-minimizable automaton 3")

    @staticmethod
    def get_nfa_1():
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state0.add_transition(zero, state0)
        state0.add_transition(zero, state1)
        state0.add_transition(one, state0)
        state1.add_transition(one, state2)
        comparator = HopcroftKarpComparison()
        return NFA(binaryAlphabet, frozenset({state0}),
                   set([state0, state1, state2]), comparator, "NFA 1")

    @staticmethod
    def get_dfa_1():
        state0 = State("State 0")
        state0_1 = State("State 0, 1")
        state0_2 = State("State 0, 2", True)
        state0.add_transition(zero, state0_1)
        state0.add_transition(one, state0)
        state0_1.add_transition(zero, state0_1)
        state0_1.add_transition(one, state0_2)
        state0_2.add_transition(zero, state0_1)
        state0_2.add_transition(one, state0)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state0_1, state0_2]), comparator, "DFA 1")

    @staticmethod
    def get_nfa_2():
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state0.add_transition(zero, state0)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state1)
        state1.add_transition(zero, state2)
        state1.add_transition(one, state1)
        state2.add_transition(zero, state2)
        state2.add_transition(zero, state1)
        state2.add_transition(one, state2)
        comparator = HopcroftKarpComparison()
        return NFA(binaryAlphabet, frozenset({state0}),
                   set([state0, state1, state2]), comparator, "NFA 2")

    @staticmethod
    def get_dfa_2():
        state0 = State("State 0")
        state1 = State("State 1")
        state1_2 = State("State 1, 2", True)
        state0.add_transition(zero, state0)
        state0.add_transition(one, state1)
        state1.add_transition(zero, state1_2)
        state1.add_transition(one, state1)
        state1_2.add_transition(zero, state1_2)
        state1_2.add_transition(one, state1_2)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state1_2]), comparator, "DFA 2")

    @staticmethod
    def get_nfa_3():
        state0 = State("State 0")
        state1 = State("State 1", True)
        state0.add_transition(zero, state0)
        state0.add_transition(zero, state1)
        state0.add_transition(one, state1)
        state1.add_transition(one, state0)
        state1.add_transition(one, state1)
        comparator = HopcroftKarpComparison()
        return NFA(binaryAlphabet, frozenset({state0}),
                   set([state0, state1]), comparator, "NFA 3")

    @staticmethod
    def get_dfa_3():
        state0 = State("State 0")
        state1 = State("State 1", True,)
        state0_1 = State("State 0, 1", True)
        state0.add_transition(zero, state0_1)
        state0.add_transition(one, state1)
        state1.add_transition(one, state0_1)
        state0_1.add_transition(zero, state0_1)
        state0_1.add_transition(one, state0_1)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, state0,
                   set([state0, state1, state0_1]), comparator, "DFA 3")

    @staticmethod
    def get_nfa_4():
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3")
        state4 = State("State 4", True)
        state5 = State("State 5")
        state6 = State("State 6", True)

        state0.add_transition(a, state0)
        state0.add_transition(a, state1)
        state0.add_transition(b, state2)
        state0.add_transition(c, state4)
        state1.add_transition(a, state4)
        state1.add_transition(a, state3)
        state1.add_transition(b, state1)
        state1.add_transition(c, state2)
        state3.add_transition(a, state3)
        state3.add_transition(a, state5)
        state5.add_transition(b, state6)
        state5.add_transition(b, state4)

        comparator = HopcroftKarpComparison()
        return NFA(abcAlphabet, frozenset({state0}),
                   set([state0, state1, state2, state3,
                        state4, state5, state6]), comparator, "NFA 4")

    @staticmethod
    def get_dfa_4():
        state1 = State("State 1")
        state4 = State("State 4", True)
        state2 = State("State 2")
        state4_3 = State("State 4, 3", True)
        state0 = State("State 0")
        state4_6 = State("State 4, 6", True)
        state3_5 = State("State 3, 5")
        state2_1 = State("State 2, 1")
        state2_4_6_1 = State("State 2, 4, 6, 1", True)
        state1_0 = State("State 1, 0")
        state4_3_1_0 = State("State 4, 3, 1, 0", True)
        state5_3_4_1_0 = State("State 5, 3, 4, 1, 0", True)
        state2_4 = State("State 2, 4", True)

        state1.add_transition(a, state4_3)
        state1.add_transition(b, state1)
        state1.add_transition(c, state2)

        state4_3.add_transition(a, state3_5)

        state0.add_transition(a, state1_0)
        state0.add_transition(b, state2)
        state0.add_transition(c, state4)

        state3_5.add_transition(a, state3_5)
        state3_5.add_transition(b, state4_6)

        state2_1.add_transition(a, state4_3)
        state2_1.add_transition(b, state1)
        state2_1.add_transition(c, state2)

        state2_4_6_1.add_transition(a, state4_3)
        state2_4_6_1.add_transition(b, state1)
        state2_4_6_1.add_transition(c, state2)

        state1_0.add_transition(a, state4_3_1_0)
        state1_0.add_transition(b, state2_1)
        state1_0.add_transition(c, state2_4)

        state4_3_1_0.add_transition(a, state5_3_4_1_0)
        state4_3_1_0.add_transition(b, state2_1)
        state4_3_1_0.add_transition(c, state2_4)

        state5_3_4_1_0.add_transition(a, state5_3_4_1_0)
        state5_3_4_1_0.add_transition(b, state2_4_6_1)
        state5_3_4_1_0.add_transition(c, state2_4)

        comparator = DFAComparator()
        return DFA(abcAlphabet, state0,
                   set([state1, state4, state2, state4_3,
                        state0, state4_6, state3_5, state2_1,
                        state2_4_6_1, state1_0, state4_3_1_0,
                        state5_3_4_1_0, state2_4]), comparator, "DFA 4")

    @staticmethod
    def get_nfa_5():
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state0.add_transition(a, state1)
        state0.add_transition(a, state2)
        state0.add_transition(b, state0)
        state0.add_transition(d, state0)
        state1.add_transition(a, state0)
        state1.add_transition(c, state2)
        state1.add_transition(d, state2)
        state1.add_transition(e, state1)
        state2.add_transition(d, state2)
        comparator = HopcroftKarpComparison()
        return NFA(abcdeAlphabet, frozenset({state0}),
                   set([state0, state1, state2]), comparator, "NFA 5")

    @staticmethod
    def get_dfa_5():
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state1_2 = State("State 1, 2", True)
        state0.add_transition(a, state1_2)
        state0.add_transition(b, state0)
        state0.add_transition(d, state0)
        state1.add_transition(a, state0)
        state1.add_transition(c, state2)
        state1.add_transition(d, state2)
        state1.add_transition(e, state1)
        state2.add_transition(d, state2)
        state1_2.add_transition(a, state0)
        state1_2.add_transition(c, state2)
        state1_2.add_transition(d, state2)
        state1_2.add_transition(e, state1)
        comparator = DFAComparator()
        return DFA(abcdeAlphabet, state0,
                   set([state0, state1, state2, state1_2]), comparator, "DFA 5")

    # Definition: Recognizes a program workflow.
    @staticmethod
    def get_program_workflow_automaton():
        stateA = State("State A")
        stateB = State("State B", True)
        stateA.add_transition(zero, stateA)
        stateA.add_transition(one, stateB)
        stateB.add_transition(zero, stateA)
        stateB.add_transition(one, stateB)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, stateA,
                   set([stateA, stateB]), comparator, "Program workflow automaton")

    # Definition: Recognizes e-commerce(taken from:
    # https://pdfs.semanticscholar.org/9cb9/74b6ece3e3fc2eab4f9cf0843bfc570df4a9.pdf).
    @staticmethod
    def get_ecommerce_automaton():
        alphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr(
            'c'), SymbolStr('d'), SymbolStr('e'), SymbolStr('f'), SymbolStr('g'), SymbolStr('h'))))
        a = alphabet['a']
        b = alphabet['b']
        c = alphabet['c']
        d = alphabet['d']
        e = alphabet['e']
        f = alphabet['f']
        g = alphabet['g']
        h = alphabet['h']
        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4", True)
        state5 = State("State 5", True)
        state6 = State("State 6", True)
        state7 = State("State 7", True)
        state0.add_transition(a, state2)
        state0.add_transition(c, state1)
        state1.add_transition(a, state3)
        state1.add_transition(c, state1)
        state2.add_transition(c, state3)
        state2.add_transition(b, state0)
        state2.add_transition(a, state2)
        state2.add_transition(d, state2)
        state2.add_transition(e, state2)
        state3.add_transition(a, state3)
        state3.add_transition(c, state3)
        state3.add_transition(d, state3)
        state3.add_transition(e, state3)
        state3.add_transition(b, state1)
        state3.add_transition(f, state4)
        state4.add_transition(b, state1)
        state4.add_transition(h, state3)
        state4.add_transition(d, state3)
        state4.add_transition(a, state3)
        state4.add_transition(f, state4)
        state4.add_transition(c, state4)
        state4.add_transition(e, state5)
        state5.add_transition(a, state6)
        state5.add_transition(d, state6)
        state5.add_transition(h, state6)
        state5.add_transition(b, state7)
        state5.add_transition(c, state5)
        state5.add_transition(e, state5)
        state5.add_transition(f, state5)
        state5.add_transition(g, state5)
        state6.add_transition(a, state6)
        state6.add_transition(c, state6)
        state6.add_transition(d, state6)
        state6.add_transition(b, state7)
        state6.add_transition(f, state5)
        state6.add_transition(g, state5)
        state7.add_transition(c, state7)
        state7.add_transition(a, state6)
        comparator = DFAComparator()
        return DFA(alphabet, state0,
                   set([state0, state1, state2, state3, state4, state5,
                        state6, state7]), comparator, "E-commerce automaton")

    # Definition: Recognizes e-commerce(taken from:
    # https://pdfs.semanticscholar.org/9cb9/74b6ece3e3fc2eab4f9cf0843bfc570df4a9.pdf).
    @staticmethod
    def get_different_ecommerce_automaton_NFA():
        alphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr(
            'c'), SymbolStr('d'), SymbolStr('e'), SymbolStr('f'), SymbolStr('g'), SymbolStr('h'))))
        a = alphabet['a']
        b = alphabet['b']
        c = alphabet['c']
        d = alphabet['d']
        e = alphabet['e']
        f = alphabet['f']
        g = alphabet['g']
        h = alphabet['h']
        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4", True)
        state5 = State("State 5", True)
        state6 = State("State 6", True)
        state7 = State("State 7", True)
        state0.add_transition(a, state2)
        state0.add_transition(c, state1)
        state1.add_transition(a, state3)
        state1.add_transition(c, state1)
        state2.add_transition(c, state3)
        state2.add_transition(b, state0)
        state2.add_transition(a, state2)
        state2.add_transition(d, state2)
        state2.add_transition(e, state2)
        state3.add_transition(a, state3)
        state3.add_transition(c, state3)
        state3.add_transition(d, state3)
        state3.add_transition(e, state3)
        state3.add_transition(c, state1)
        state3.add_transition(f, state4)
        state4.add_transition(b, state1)
        state4.add_transition(h, state3)
        state4.add_transition(d, state3)
        state4.add_transition(a, state3)
        state4.add_transition(f, state4)
        state4.add_transition(c, state4)
        state4.add_transition(e, state5)
        state5.add_transition(d, state6)
        state5.add_transition(h, state6)
        state5.add_transition(b, state7)
        state5.add_transition(c, state5)
        state5.add_transition(e, state5)
        state5.add_transition(f, state5)
        state5.add_transition(g, state5)
        state6.add_transition(b, state6)
        state6.add_transition(c, state6)
        state6.add_transition(b, state7)
        state6.add_transition(f, state5)
        state6.add_transition(h, state5)
        state7.add_transition(c, state7)
        state7.add_transition(a, state6)
        comparator = HopcroftKarpComparison()
        return NFA(alphabet, frozenset({state0}),
                   set([state0, state1, state2, state3, state4, state5,
                        state6, state7]), comparator, "Modified E-commerce automaton")

    # Definition: Recognizes e-commerce, reduced version (taken from:
    # https://pdfs.semanticscholar.org/9cb9/74b6ece3e3fc2eab4f9cf0843bfc570df4a9.pdf).
    @staticmethod
    def get_reduced_ecommerce_automaton():
        alphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr(
            'c'), SymbolStr('d'), SymbolStr('e'), SymbolStr('f'), SymbolStr('h'))))
        a = alphabet['a']
        b = alphabet['b']
        c = alphabet['c']
        d = alphabet['d']
        e = alphabet['e']
        f = alphabet['f']
        h = alphabet['h']
        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4", True)
        state0.add_transition(a, state2)
        state0.add_transition(c, state1)
        state1.add_transition(a, state3)
        state1.add_transition(c, state1)
        state2.add_transition(c, state3)
        state2.add_transition(b, state0)
        state2.add_transition(a, state2)
        state2.add_transition(d, state2)
        state2.add_transition(e, state2)
        state3.add_transition(a, state3)
        state3.add_transition(c, state3)
        state3.add_transition(d, state3)
        state3.add_transition(e, state3)
        state3.add_transition(b, state1)
        state3.add_transition(f, state4)
        state4.add_transition(b, state1)
        state4.add_transition(h, state3)
        state4.add_transition(d, state3)
        state4.add_transition(a, state3)
        state4.add_transition(f, state4)
        state4.add_transition(c, state4)
        state4.add_transition(e, state4)
        comparator = DFAComparator()
        return DFA(alphabet, state0,
                   set([state0, state1, state2, state3, state4]),
                   comparator,
                   "Reduced E-commerce automaton")

    # Definition: Recognizes strings not containing 'bbb' (not minimized).
    @staticmethod
    def get_dolzhenko_jonoska_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        state0 = State("State 0", True)
        state1 = State("State 1", True)
        state2 = State("State 2", True)
        state3 = State("State 3", True)
        state4 = State("State 4", True)
        state5 = State("State 5", True)
        state6 = State("State 6", True)
        state7 = State("State 7", True)
        state0.add_transition(a, state1)
        state0.add_transition(b, state4)
        state1.add_transition(a, state1)
        state1.add_transition(b, state2)
        state2.add_transition(a, state1)
        state2.add_transition(b, state3)
        state3.add_transition(a, state1)
        state4.add_transition(a, state6)
        state4.add_transition(b, state5)
        state5.add_transition(a, state6)
        state6.add_transition(a, state6)
        state6.add_transition(b, state7)
        state7.add_transition(a, state6)
        state7.add_transition(b, state5)
        comparator = DFAComparator()
        return DFA(abAlphabet, state0,
                   set([state0, state1, state2, state3, state4, state5,
                        state6, state7]), comparator,
                   "Dolzhenko-Jonoska automaton")

    # Definition: (a|b)*
    @staticmethod
    def get_a_or_b_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateA = State("State A", True)
        stateA.add_transition(a, stateA)
        stateA.add_transition(b, stateA)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateA, set([stateA]), comparator, "a or b Automaton")

    # Definition: (c*ac*bc*)*
    @staticmethod
    def get_ab_with_cs_automaton():
        a = abcAlphabet['a']
        b = abcAlphabet['b']
        c = abcAlphabet['c']
        state0 = State("State 0", True)
        state1 = State("State 1")
        state0.add_transition(a, state1)
        state0.add_transition(c, state0)
        state1.add_transition(b, state0)
        state1.add_transition(c, state1)
        comparator = DFAComparator()
        return DFA(abcAlphabet, state0,
                   set([state0, state1]), comparator, "ab with cs automaton")

    # Definition: (ab)*
    @staticmethod
    def get_ab_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        state0 = State("State 0", True)
        state1 = State("State 1")
        state0.add_transition(a, state1)
        state1.add_transition(b, state0)
        comparator = DFAComparator()
        return DFA(abAlphabet, state0,
                   set([state0, state1]), comparator, "ab automaton")

    # Definition: (.*0)
    @staticmethod
    def get_zero_ending_automaton():
        stateA = State("State A", True)
        stateB = State("State B", True)
        stateC = State("State C")
        stateA.add_transition(zero, stateA)
        stateA.add_transition(one, stateC)
        stateB.add_transition(zero, stateA)
        stateB.add_transition(one, stateC)
        stateC.add_transition(zero, stateA)
        stateC.add_transition(one, stateC)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, stateA,
                   set([stateA, stateB, stateC]), comparator, "Zero ending automaton")

    # Definition: (.*a), with {a, b, c} alphabet.
    @staticmethod
    def get_a_ending_with_cs_automaton():
        a = abcAlphabet['a']
        b = abcAlphabet['b']
        c = abcAlphabet['c']
        state0 = State("State 0", True)
        state1 = State("State 1")
        state0.add_transition(a, state0)
        state0.add_transition(b, state1)
        state0.add_transition(c, state0)
        state1.add_transition(a, state0)
        state1.add_transition(b, state1)
        state1.add_transition(c, state1)
        comparator = DFAComparator()
        return DFA(abcAlphabet, state0,
                   set([state0, state1]), comparator, "a ending with cs automaton")

    # Definition: (.*a), with {a, b} alphabet.
    @staticmethod
    def get_a_ending_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        state0 = State("State 0", True)
        state1 = State("State 1")
        state0.add_transition(a, state0)
        state0.add_transition(b, state1)
        state1.add_transition(a, state0)
        state1.add_transition(b, state1)
        comparator = DFAComparator()
        return DFA(abAlphabet, state0,
                   set([state0, state1]), comparator, "a ending automaton")

    # Definition: (..*0)
    @staticmethod
    def get_nonempty_zero_ending_automaton():
        stateA = State("State A")
        stateB = State("State B", True)
        stateA.add_transition(zero, stateB)
        stateA.add_transition(one, stateA)
        stateB.add_transition(zero, stateB)
        stateB.add_transition(one, stateA)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, stateA,
                   set([stateA, stateB]), comparator, "Non-empty zero ending automaton")

    # Definition: (..*1)
    @staticmethod
    def get_nonempty_one_ending_automaton():
        stateA = State("State A")
        stateB = State("State B", True)
        stateA.add_transition(zero, stateA)
        stateA.add_transition(one, stateB)
        stateB.add_transition(zero, stateA)
        stateB.add_transition(one, stateB)
        comparator = DFAComparator()
        return DFA(binaryAlphabet, stateA,
                   set([stateA, stateB]), comparator, "Non-empty one ending automaton")

    # Definition: (ab)*a
    @staticmethod
    def get_complex_ab_prefixed_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state3 = State("State 3")
        state4 = State("State 4")
        state5 = State("State 5")
        state6 = State("State 6")
        state7 = State("State 7")
        state0.add_transition(a, state1)
        state0.add_transition(b, state5)
        state1.add_transition(a, state6)
        state1.add_transition(b, state2)
        state2.add_transition(a, state0)
        state2.add_transition(b, state2)
        state3.add_transition(a, state2)
        state3.add_transition(b, state6)
        state4.add_transition(a, state7)
        state4.add_transition(b, state5)
        state5.add_transition(a, state2)
        state5.add_transition(b, state6)
        state6.add_transition(a, state6)
        state6.add_transition(b, state4)
        state7.add_transition(a, state6)
        state7.add_transition(b, state2)
        comparator = DFAComparator()
        return DFA(abAlphabet, state0,
                   set([state0, state1, state2, state3, state4, state5,
                        state6, state7]), comparator, "Complex ab prefixed automaton")

    # Definition: (ab)*a
    @staticmethod
    def get_simpler_ab_prefixed_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        state0 = State("State 0")
        state1 = State("State 1", True)
        state0.add_transition(a, state1)
        state1.add_transition(b, state0)
        comparator = DFAComparator()
        return DFA(abAlphabet, state0,
                   set([state0, state1]), comparator, "Simpler ab prefixed automaton")

    @staticmethod
    def get_alternating_bit_protocol_automaton():
        a = abcdAlphabet['a']
        b = abcdAlphabet['b']
        c = abcdAlphabet['c']
        d = abcdAlphabet['d']
        state0 = State("State 0", True)
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3")
        state0.add_transition(a, state0)
        state0.add_transition(d, state1)
        state1.add_transition(d, state1)
        state1.add_transition(b, state2)
        state2.add_transition(b, state2)
        state2.add_transition(c, state3)
        state3.add_transition(c, state3)
        state3.add_transition(a, state0)
        comparator = DFAComparator()
        return DFA(abcdAlphabet, state0,
                   set([state0, state1, state2, state3]),
                   comparator,
                   "Alternating bit protocol automaton")

    @staticmethod
    def get_alternating_bit_protocol_z_automaton():
        alphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'),
                                       SymbolStr('c'), SymbolStr('d'), SymbolStr('e'))))
        a = alphabet['a']
        b = alphabet['b']
        c = alphabet['c']
        d = alphabet['d']
        e = alphabet['e']
        state0 = State("State 0", True)
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3")
        state0.add_transition(a, state0)
        state0.add_transition(d, state1)
        state0.add_transition(e, state0)
        state1.add_transition(d, state1)
        state1.add_transition(b, state2)
        state1.add_transition(e, state1)
        state2.add_transition(b, state2)
        state2.add_transition(c, state3)
        state2.add_transition(e, state2)
        state3.add_transition(c, state3)
        state3.add_transition(a, state0)
        state3.add_transition(e, state3)
        comparator = DFAComparator()
        return DFA(alphabet, state0,
                   set([state0, state1, state2, state3]),
                   comparator,
                   "Alternating bit protocol z automaton")

    # Definition: {}
    @staticmethod
    def get_empty_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateA = State("State A", False)
        stateA.add_transition(a, stateA)
        stateA.add_transition(b, stateA)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateA,
                   set([stateA]), comparator, "Empty automaton")

    # Definition: Σ*
    @staticmethod
    def get_sigma_star_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateA = State("State A", True)
        stateA.add_transition(a, stateA)
        stateA.add_transition(b, stateA)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateA,
                   set([stateA]), comparator, "Sigma-star automaton")

    # Definition: uneven number of as
    @staticmethod
    def get_uneven_number_of_as_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateA = State("State A", False)
        stateB = State("State B", True)
        stateA.add_transition(a, stateB)
        stateA.add_transition(b, stateA)
        stateB.add_transition(a, stateA)
        stateB.add_transition(b, stateB)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateA,
                   set([stateA, stateB]), comparator, "Uneven number of as")

    # Definition: uneven number of symbols
    @staticmethod
    def get_uneven_number_of_symbols_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateC = State("State C", False)
        stateD = State("State D", True)
        stateC.add_transition(a, stateD)
        stateC.add_transition(b, stateD)
        stateD.add_transition(a, stateC)
        stateD.add_transition(b, stateC)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateC,
                   set([stateC, stateD]), comparator, "Uneven number of symbols")

    # Definition: uneven number of as and symbols
    @staticmethod
    def get_uneven_number_of_as_and_symbols_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateAC = State("State AC", False)
        stateAD = State("State AD", False)
        stateBC = State("State BC", False)
        stateBD = State("State BD", True)

        stateAC.add_transition(a, stateBD)
        stateAC.add_transition(b, stateAD)
        stateAD.add_transition(a, stateBC)
        stateAD.add_transition(b, stateAC)
        stateBC.add_transition(a, stateAD)
        stateBC.add_transition(b, stateBD)
        stateBD.add_transition(a, stateAC)
        stateBD.add_transition(b, stateBC)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateAC,
                   set([stateAC, stateAD, stateBC, stateBD]),
                   comparator,
                   "Uneven number of as and symbols")

    # Definition: uneven number of as or symbols
    @staticmethod
    def get_uneven_number_of_as_or_symbols_automaton():
        a = abAlphabet['a']
        b = abAlphabet['b']
        stateAC = State("State AC", False)
        stateAD = State("State AD", True)
        stateBC = State("State BC", True)
        stateBD = State("State BD", True)

        stateAC.add_transition(a, stateBD)
        stateAC.add_transition(b, stateAD)
        stateAD.add_transition(a, stateBC)
        stateAD.add_transition(b, stateAC)
        stateBC.add_transition(a, stateAD)
        stateBC.add_transition(b, stateBD)
        stateBD.add_transition(a, stateAC)
        stateBD.add_transition(b, stateBC)
        comparator = DFAComparator()
        return DFA(abAlphabet, stateAC,
                   set([stateAC, stateAD, stateBC, stateBD]),
                   comparator,
                   "Uneven number of as or symbols")



from graphviz import Digraph
from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy


class ImageExportingStrategyWithoutHoleState(ModelExportingStrategy):

    '''
    Exports pdf with automaton image not plotting the hole state.
    Hole state should be assigned, otherwise if there is a real hole state it will still be plotted.
    '''
    def export(self, model, path=None):   
        graph = self._create_graph(model)
        path = self.get_path_for(path, model)
        graph.render(path, cleanup=True)

    def _create_graph(self, model):
        graph = Digraph('finite_state_machine')
        graph.attr(rankdir='LR', size='8,5')

        finalStates = filter(lambda state: state.is_final, model.states)
        graph.attr('node', shape='doublecircle')
        
        model_has_one_state = len(model.states)==1
        for state in finalStates:
            if state != model.hole or model_has_one_state:
                graph.node(state.name)

        graph.attr('node', shape='circle')
        for state in model.states:
            transitions = dict()
            for symbol, destinationStates in state.transitions.items():
                for destinationState in destinationStates:
                    if destinationState != model.hole or model_has_one_state:
                        transitions.setdefault((state.name, destinationState.name), set())
                        transitions[(state.name, destinationState.name)].add(str(symbol))
            for key, symbols in transitions.items():
                state_from, state_to = key
                label = self._get_label_for(symbols, model.alphabet)
                graph.edge(state_from, state_to, label)

        graph.attr('node', shape='point')
        for index, initialState in enumerate(model.initial_states):
            nodeName = 'start' + str(index)
            graph.node(nodeName)
            graph.edge(nodeName, initialState.name)
        
        return graph

    def _get_label_for(self, symbols: list, alphabet):
        if len(symbols) == len(alphabet):
            return "Σ"
        elif len(symbols) <= len(alphabet) / 2:
            return ", ".join(sorted(symbols))
        else:
            complement = set(map(str, alphabet.symbols)) - set(symbols)
            label = ", ".join(sorted(complement))
            return f"Σ - {{{label}}}"



from collections import abc
from pythautomata.base_types.symbol import Symbol
from typing import Union
from pythautomata.exceptions.unexpected_type_exception import UnexpectedTypeException


class Sequence():
    """Ordered collection of Symbols.
    """

    def __init__(self, value: Union['Sequence', list[Symbol], tuple[Symbol]] = ()):
        if value == ():
            self._value: tuple[Symbol] = ()
        elif isinstance(value, Sequence):
            self._value = value.value
        elif isinstance(value, list):
            self._value = tuple(value)
        elif isinstance(value, tuple):
            self._value = value
        else:
            raise ValueError

    @property
    def value(self) -> tuple[Symbol]:
        return self._value

    def get_prefixes(self) -> list['Sequence']:
        result = list()
        for i in range(1, len(self.value) + 1):
            result.append(Sequence(self.value[:i]))
        return result

    def get_suffixes(self) -> list['Sequence']:
        result = list()
        for i in range(0, len(self.value)):
            result.append(Sequence(self.value[i:]))
        return result

    def append(self, symbol_to_append: Symbol):
        value = list(self.value)
        value.append(symbol_to_append)
        return Sequence(value)

    def __getitem__(self, key):
        return self.value[key]

    def __len__(self):
        return len(self.value)

    def __add__(self, other: Union['Sequence', Symbol]) -> 'Sequence':
        if isinstance(other, Sequence):
            return Sequence(self.value + other.value)
        if isinstance(other, Symbol):
            return Sequence(self.value + (other,))
        raise UnexpectedTypeException()

    def __radd__(self, other: Union['Sequence', Symbol]) -> 'Sequence':
        if isinstance(other, Sequence):
            return Sequence(other.value + self.value)
        if isinstance(other, Symbol):
            return Sequence((other,) + self.value)
        raise UnexpectedTypeException()

    def __repr__(self):
        return "ϵ" if self.value == () else "".join(map(lambda x: str(x), self.value))

    def __lt__(self, other: 'Sequence') -> bool:
        if len(self.value) == len(other.value):
            return self.__repr__() < other.__repr__()
        else:
            return len(self.value) < len(other.value)

    def __iter__(self):
        return self.value.__iter__()

    def __eq__(self, other) -> bool:
        if isinstance(other, Sequence):
            return self.value == other.value
        return False

    def __hash__(self):
        return hash(self._value)



from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy
from pythautomata.automata.non_deterministic_finite_automaton import NondeterministicFiniteAutomaton
from pythautomata.base_types.symbolic_state import SymbolicState
from graphviz import Digraph


class DotExportingStrategy(ModelExportingStrategy):

    def export(self, model, path=None):
        graph = self.create_graph(model)
        path = self.get_path_for(path, model)
        path = str(path) + '.dot'

        dot_code = graph.source

        with open(path, "w+", encoding="utf-8") as f:
            f.write(dot_code)

    def create_graph(self, model):
        graph = Digraph(model.name)
        graph.attr(rankdir='LR', size='8,5')

        finalStates = filter(lambda state: state.is_final, model.states)
        graph.attr('node', shape='doublecircle')
        for state in finalStates:
            graph.node(state.name)

        graph.attr('node', shape='circle')
        for state in model.states:
            transitions = dict()
            if type(state) == SymbolicState:
                for symbol, destinationState in state.transitions:
                    transitions.setdefault(
                        (state.name, destinationState.name), set())
                    transitions[(state.name, destinationState.name)].add(
                        str(symbol))
            else:
                for symbol, destinationStates in state.transitions.items():
                    for destinationState in destinationStates:
                        transitions.setdefault(
                            (state.name, destinationState.name), set())
                        transitions[(state.name, destinationState.name)].add(
                            str(symbol))

            for key, symbols in transitions.items():
                state_from, state_to = key
                label = self.get_label_for(symbols, model.alphabet)
                graph.edge(state_from, state_to, label)

        graph.attr('node', shape='point')
        if type(model) == NondeterministicFiniteAutomaton:
            for index, initialState in enumerate(model.initial_states):
                nodeName = 'start' + str(index)
                graph.node(nodeName)
                graph.edge(nodeName, initialState.name)
        else:
            initialState = model.initial_state
            nodeName = 'start'
            graph.node(nodeName)
            graph.edge(nodeName, initialState.name)
        return graph

    def get_label_for(self, symbols: list, alphabet):
        if len(symbols) == len(alphabet):
            return "Σ"
        elif len(symbols) <= len(alphabet) / 2:
            return ", ".join(sorted(symbols))
        else:
            complement = set(map(str, alphabet.symbols)) - set(symbols)
            label = ", ".join(sorted(complement))
            return f"Σ - {{{label}}}"



from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.guard import Guard


class AndGuard(Guard):
    def __init__(self, guard1: Guard, guard2: Guard):
        self.guard1 = guard1
        self.guard2 = guard2

    def matches(self, s: Symbol):
        return self.guard1.matches(s) and self.guard2.matches(s)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "(" + str(self.guard1) + u") \u2227 (" + str(self.guard2) + ")"



from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy


class EncodedStringExportingStrategy(ModelExportingStrategy):

    def export(self, model):
        my_string = ""
        my_string += model.name+"\n"
        my_string += "Alphabet: ["
        my_string += ", ".join(
            map(lambda symbol: f"'{str(symbol)}'", sorted(model.alphabet.symbols)))

        my_string += "]\n\nStates:"
        sortedStates = sorted(model.states, key=lambda state: state.name)
        for state in sortedStates:
            my_string += f"\n\t\"{state.name}\""
            if state in model.initial_states:
                my_string += " (Initial)"
            if state.is_final:
                my_string += " (Final)"

        my_string += "\n\nTransitions:\n\n"
        my_string += "\n".join(map(self._get_all_transitions_lines_for,
                                   sortedStates))
        return my_string

    def _get_all_transitions_lines_for(self, state):
        if not state.transitions:
            return ""
        sortedTransitions = sorted(state.transitions.items())
        transitionsLines = "\n\t".join(
            map(self._get_one_symbol_transitions_lines_for,
                sortedTransitions))
        return f"\"{state.name}\":\n\t{transitionsLines}"

    def _get_one_symbol_transitions_lines_for(self, item):
        substring = ", ".join(
            map(lambda next_state: f"\"{next_state.name}\"", item[1]))
        return f"'{str(item[0])}' -> [{substring}]"
